<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>Carrot</title>

    <link rel="icon" type="image/png" href="{{ url_for('static', filename='img/carrot.png') }}" />

    <link href="https://fonts.googleapis.com/css?family=Roboto|Source+Code+Pro" rel="stylesheet">

    <link
            rel="stylesheet"
            href="https://use.fontawesome.com/releases/v5.6.3/css/all.css"
            integrity="sha384-UHRtZLI+pbxtHCWp1t77Bi1L4ZtiqrqD80Kn4Z8NTSRyMA2Fd33n5dQ8lWUE00s/"
            crossorigin="anonymous">

    <link href="/static/main.css" rel="stylesheet">
</head>
<body>
    <div id="root"></div>

    <script type="text/javascript" src="/static/main.js"></script>
</body>
</html>